/**
 * \file LazyWorkflow.cpp
 *
 *  \date Apr 12, 2012
 *  \author Alexandre Tuleu
 */

#include "LazyWorkflow.h"

#include <libsbcp/errors/Errors.h>
#include "Bus.h"

#include <iostream>

using namespace sbcp;

const unsigned int LazyWorkflow::WORKFLOW(Bus::LAZY);

LazyWorkflow::LazyWorkflow(Bus & bus)
	: BusWorkflow(bus,WORKFLOW)
	, d_transfer(NULL){
	sbcpd_bus_release_all_transfers(d_bus.SBCPD());
	d_transfer = NULL;
}

LazyWorkflow::~LazyWorkflow(){
	if(d_transfer != NULL){
		this->Release();
	}
}


Byte * LazyWorkflow::Prepare(std::tr1::shared_ptr<Device> dev ,
                                     InstructionID instruction , PayloadSize size){

	return Prepare(dev->Klass(),dev->ID(),instruction,size);
}



Byte * LazyWorkflow::Prepare(sbcp::Class Klass , DevID id ,
                             InstructionID instruction , PayloadSize size){
	if(d_transfer != NULL){
		throw std::runtime_error ("could only perform one transfer at time, please Release() operation.");
	}

	if(!d_bus.Good()){
		std::runtime_error ("bus is not ready for io operation.");
	}

	d_transfer = sbcpd_bus_append_transfer(d_bus.SBCPD(),size);

	if(!d_transfer){
		throw SbcdpError("transfer initialiazation",d_bus.SBCPD());
	}

	d_transfer->callback = NULL;
	d_transfer->out_buffer[SBCP_CLASS_POS]        = Klass;
	d_transfer->out_buffer[SBCP_ID_POS]           = id;
	d_transfer->out_buffer[SBCP_INSTRUCTION_POS]  = instruction;
	d_transfer->out_buffer[SBCP_PAYLOAD_SIZE_POS] = size;
	return &(d_transfer->out_buffer[SBCP_PAYLOAD_START_POS]);


}

void LazyWorkflow::Submit(const Byte** data , PayloadSize& size){
	if(d_transfer == NULL){
		throw std::runtime_error("Lazy workflow error : no transfer prepared with Prepare()");
	}
	if(sbcpd_bus_submit_all_transfer(d_bus.SBCPD()) != SBCPD_NO_ERROR){
		throw SbcdpError("submission",d_bus.SBCPD());
	}

	if(sbcpd_bus_wait_for_completion(d_bus.SBCPD(),-1) != SBCPD_NO_ERROR){
		throw SbcdpError("wait for completion", d_bus.SBCPD());
	}

	*data = &(d_transfer->in_buffer[SBCP_PAYLOAD_START_POS]);
	size = d_transfer->in_size - SBCP_MINIMAL_SIZE;

	if(d_transfer->status != SBCPD_TRANSFER_OK){
		TransferError e(d_transfer);
		Release();
		throw e;
	}

	if( (d_transfer->in_buffer[SBCP_CLASS_POS] != d_transfer->out_buffer[SBCP_CLASS_POS]) ||
	    (d_transfer->in_buffer[SBCP_ID_POS]    != d_transfer->out_buffer[SBCP_ID_POS])){
		TransferError e(d_transfer);
		Release();
		throw e;
	}

}

void LazyWorkflow::Release(){
	if(d_transfer == NULL){
		throw std::runtime_error("No transfer prepared.");
	}

	sbcpd_bus_release_all_transfers(d_bus.SBCPD());
	d_transfer = NULL;
}

BusWorkflow::Ptr LazyWorkflow::Creator::Create(Bus & bus){
	return BusWorkflow::Ptr(new LazyWorkflow(bus));
}



