/**
 * \file ScheduledWorkflow.cpp
 *
 *  \date Apr 12, 2012
 *  \author Alexandre Tuleu
 */

#include "ScheduledWorkflow.h"

#include <stdexcept>
#include <iostream>
#include <sstream>

#include <sbcpd/transfer.h>

#include "Bus.h"


namespace sbcp {
	const unsigned int ScheduledWorkflow::WORKFLOW(Bus::SCHEDULED);

	ScheduledWorkflow::ScheduledWorkflow(Bus & bus)
		: BusWorkflow(bus,WORKFLOW)
		, d_state(PREPARATION)
		, d_transfers(){
	}

	ScheduledWorkflow::~ScheduledWorkflow(){
		WaitForTransfersCompletion();
		ClearAllScheduledDevices();
	}


	void ScheduledWorkflow::AppendScheduledDevice(Device::Ptr device){
		if(d_state != PREPARATION){
			throw std::runtime_error("Could not schedule a new device after first transmission, please ClearAllScheduledDevices() first");
		}


		if(&d_bus != &(device->GetBus())){
			throw std::runtime_error("Device should be on the same bus !");
		}

		device->CheckSchedulable();

		TransferByDevice::const_iterator fi = d_transfers.find(device);

		if( fi != d_transfers.end() ){
			return;
		}

		sbcpd_transfer * t = sbcpd_bus_append_transfer(d_bus.SBCPD(),device->LowLatencyOutSize());

		if (t == NULL ) {
			throw std::runtime_error(std::string("sbcpd error on transfer initialization : ") +
			                         sbcpd_bus_get_last_error_message(d_bus.SBCPD()));
		}

		t->out_buffer[SBCP_CLASS_POS] = device->Klass();
		t->out_buffer[SBCP_ID_POS]    = device->ID();
		t->out_buffer[SBCP_INSTRUCTION_POS] = Device::LOW_LATENCY_INSTRUCTION;
		t->out_buffer[SBCP_PAYLOAD_SIZE_POS] = device->LowLatencyOutSize();
		d_transfers.insert(std::make_pair(device,t));

	}

	void ScheduledWorkflow::ClearAllScheduledDevices(){
		if(d_state == WAITING_FOR_COMPLETION){
			throw std::runtime_error("Transfer already submitted, please WaitForTransfersCompletion() first.");
		}

		ReleaseAllIOTransfers();
		d_transfers.clear();

		d_state = PREPARATION;

	}

	void ScheduledWorkflow::StartTransfers(){
		if(d_state == WAITING_FOR_COMPLETION){
			throw std::runtime_error("Transfers already started!");
		}


		for(TransferByDevice::iterator t = d_transfers.begin();
		    t != d_transfers.end();
		    ++t){

			Device::LowLatencyIOAccessor::WriteLowLatencyBuffer(*t->first,
			                                                    &(t->second->out_buffer[SBCP_PAYLOAD_START_POS]),
			                                                    t->second->out_buffer[SBCP_PAYLOAD_SIZE_POS]);
		}

		if( sbcpd_bus_submit_all_transfer(d_bus.SBCPD()) != SBCPD_NO_ERROR ){
					throw std::runtime_error(std::string("sbcpd bus error on submission : ") +
					                         sbcpd_bus_get_last_error_message(d_bus.SBCPD()));

		}


		d_state = WAITING_FOR_COMPLETION;
	}

	void ScheduledWorkflow::WaitForTransfersCompletion(){
		if(sbcpd_bus_wait_for_completion(d_bus.SBCPD(),-1) != SBCPD_NO_ERROR){
			throw std::runtime_error(std::string("sbcpd bus error on completion : ") +
			                         sbcpd_bus_get_last_error_message(d_bus.SBCPD()));
		}

		if(d_state != WAITING_FOR_COMPLETION ){
			return;
		}

		MultipleTransferError e;
		for(TransferByDevice::iterator t = d_transfers.begin();
		    t != d_transfers.end();
		    ++t){
			if(t->second->status != SBCPD_TRANSFER_OK){
				e.AppendError(TransferError(t->second));
				continue;
			}
			if((t->second->out_buffer[SBCP_CLASS_POS] != t->second->in_buffer[SBCP_CLASS_POS]) ||
			   (t->second->out_buffer[SBCP_ID_POS]    != t->second->in_buffer[SBCP_ID_POS])){
				e.AppendError(TransferError(t->second));
			}

			Device::LowLatencyIOAccessor::ReadLowLatencyBuffer(*(t->first),
			                                                   &(t->second->in_buffer[SBCP_PAYLOAD_START_POS]),
			                                                   t->second->in_buffer[SBCP_PAYLOAD_SIZE_POS]);
		}

		d_state = WAITING_FOR_RESUBMISSION;

		if(! e.Errors().empty()){
			throw e;
		}


	}

	void ScheduledWorkflow::ReleaseAllIOTransfers(){
		if(sbcpd_bus_release_all_transfers(d_bus.SBCPD()) != SBCPD_NO_ERROR){
			throw std::runtime_error(std::string("sbcpd bus error on release : ") +
			                         sbcpd_bus_get_last_error_message(d_bus.SBCPD()));

		}
	}

	BusWorkflow::Ptr ScheduledWorkflow::Creator::Create(Bus & bus){
		return BusWorkflow::Ptr(new ScheduledWorkflow(bus));
	}

} /* namespace sbcp */
