/**
 * \file device/Device.cpp
 *
 *  \date Mar 31, 2012
 *  \author Alexandre Tuleu
 */

#include "Device.h"

#include <libsbcp/bus/Bus.h>
#include <cstring>
#include <stdexcept>
#include <sstream>
#include <iomanip>

#include <iostream>

namespace sbcp {



	Device::Device(Bus & bus,
	               sbcp::Class klass ,
	               DevID id,
	               FirmwareVersion version,
	               SBCPVersion comVersion)
	: d_class(klass)
	, d_id(id)
	, d_version(version)
	, d_comVersion(comVersion)
	, d_registers()
	, d_bus(bus)
	, d_llReaders()
	, d_llWriters()
	, d_maxllInputSize(0)
	, d_maxllOutputSize(0)
	, d_fillingStdRegister(false){
		//reserve in the vector for the maximal size of register in the device.
		FillStdRegisters();
	}

	Device::~Device(){
	}

	Device::ListOfRegisters & Device::Registers(){
		return d_registers;
	}

	sbcp::Class Device::Klass() const{
		return d_class;
	}

	DevID Device::ID() const {
		return d_id;
	}

	FirmwareVersion Device::Version() const {
		return d_version;
	}

#define CHECK_ACCES_AND_THROW(a)do{\
	if(a >= d_registers.size()){\
		std::ostringstream os;\
		os << "0x" << std::hex << (int) address\
		   << " is out of bound; device has " << d_registers.size()\
		   << " registers." << std::flush;\
		throw std::overflow_error(os.str());\
	}\
}while(0)


#define ADD_DEVICE_HEADER(dev,os,reg)do{	  \
	os << "Device 0x" << std::hex\
	   << (int) ((Device*) dev )->Klass() << ":0x"\
	   << (int) ((Device*) dev )->ID() << std::dec\
	   << " Register " << (int) reg;\
}while(0)

	void Device::SetRegister(Register::Address address, Data data){
//		CHECK_ACCES_AND_THROW(address);

		if(!(d_registers[address]->Flags() & Register::WRITABLE)){
			throw std::logic_error("Register is not writable.");
			return;
		}

		if(d_bus.CurrentWorkflow() == Bus::SCHEDULED){
			throw std::runtime_error("The workflow is scheduled, could not have lazy IO.");
		}

		//switch to lazy workflow.
		LazyWorkflow & w = d_bus.Lazy();

		Byte * out = w.Prepare(Klass(),ID(),SET_REGISTERS,3);
		out[0] = address;
		FillBytesFromData(data,&(out[1]));

		const Byte * res(NULL);
		PayloadSize resSize;

		w.Submit(&res,resSize);


		if(resSize != 0 ){
			std::ostringstream os;
			ADD_DEVICE_HEADER(this,os,address);
			os << " received a payload of " << resSize << " for set instruction but none expected.";
			throw std::runtime_error(os.str());
		}

		w.Release();
	}

	Data Device::GetRegister(Register::Address address ){
//		CHECK_ACCES_AND_THROW(address);
		Register * reg = d_registers[address];
		Data d;
		if(!(reg->Flags() & Register::READABLE)){
			throw std::logic_error("The register is not readable.");
		}
		if(d_bus.CurrentWorkflow() == Bus::SCHEDULED){
			throw std::runtime_error("The workflow is scheduled, could not have lazy IO.");
		}

		//switch to LazyWorkflow
		LazyWorkflow & w = d_bus.Lazy();
		Byte * out = w.Prepare(Klass(),ID(), GET_REGISTERS,1);
		out[0] = address;
		const Byte * res(NULL);
		PayloadSize resSize;
		w.Submit(&res,resSize);

		if( resSize != 2){
			std::ostringstream os;
			ADD_DEVICE_HEADER(this,os,address);
			os << " got " << resSize << " bytes but 2 is expected to get a register";
			throw std::runtime_error(os.str());
		}


		FillDataFromBytes(res,d);
		w.Release();
		return d;
	}

	SBCPVersion Device::ComVersion() const{
		return d_comVersion;
	}

	const Device::ListOfWriters& Device::GetLowLatencyInPacket() const{
		return d_llWriters;
	}

	const Device::ListOfReaders& Device::GetLowLatencyOutPacket() const{
		return d_llReaders;
	}

	bool Device::HasALowLatencyPacketDefinition(){
		return ! (d_llReaders.empty() && d_llWriters.empty());
	}

	bool Device::CanHaveALowLatencyPacket(){
		return d_maxllInputSize != 0 || d_maxllOutputSize != 0;
	}

	void Device::FillStdRegisters(){
		typedef std::tr1::shared_ptr<Register> RPtr;
		d_registers.resize(sbcp::MaxDevRegisters);

		d_fillingStdRegister = true;
		d_firmwareReg = RPtr(new Register(*this,
		                                  FIRMWARE_VERSION,
		                                  Register::READABLE | Register::PERSISTENT));
		d_classReg = RPtr(new Register(*this,
		                               SBCP_CLASS,
		                               Register::READABLE | Register::PERSISTENT));
		d_firmwareReg = RPtr(new Register(*this,
		                                  SBCP_ID,
		                                  Register::WRITABLE | Register::READABLE | Register::PERSISTENT));
		d_fillingStdRegister = false;
	}


	void Device::AddRegister(Register * reg){
		// DevRegister::Address defintion insure address < MaxDevRegisters
		// this is an uint8_t !
		Register::Address address = reg->GetAddress();
		if(address < DEVICE_SPECIFIC_REGISTERS && !d_fillingStdRegister){
			std::ostringstream os;
			os << std::hex << "Adress 0x" << (int) address
			   << " is reserved for standard registers please "
			   << " choose an address >= 0x" << (int) DEVICE_SPECIFIC_REGISTERS
			   << std::endl;
			throw std::underflow_error(os.str());
		}

		if(d_registers.size() > address) {
			if(d_registers[address] != NULL) {
				std::ostringstream os;
				os << std::hex << "Address 0x" << (int) address
				   << " is already in use in this device";
				throw std::logic_error(os.str());
			}
			d_registers[address] = reg;
		} else {
			//zero allocate the memory
			d_registers.resize((ListOfRegisters::size_type) address + 1, NULL);
			d_registers[address] = reg;
		}

	}


	Bus & Device::GetBus() {
		return d_bus;
	}

	Register::Flag
	Device::GetRegisterFlags(Register::Address address){
//		CHECK_ACCES_AND_THROW(address);
		return (Register::Flag) (d_registers[address]->Flags() & Register::PublicFlagsMask);
	}




	void Device::WriteLowLatencyBuffer(Byte* buffer , PayloadSize size){
		PayloadSize expectedSize = LowLatencyOutSize();
		if(size != expectedSize){
			std::ostringstream os;
			os << "expecting a transfer output of " << (int) expectedSize
			   << " but got a buffer size of " << (int)size << std::endl;
			throw std::logic_error(os.str());
		}

		for(unsigned int i = 0; i < d_llWriters.size() ; ++i){
			FillBytesFromData(d_llWriters[i]->ToData(),&(buffer[2*i]));
		}
	}

	void Device::ReadLowLatencyBuffer(const Byte* buffer , PayloadSize size){
		PayloadSize expectedSize = LowLatencyInSize();
		if(size != expectedSize){
			std::ostringstream os;
			os << "expecting a transfer input of " << (int) expectedSize
			   << " but got a buffer size of " << (int)size << std::endl;
			throw std::runtime_error(os.str());
		}

		Data d;
		for(unsigned int i = 0; i < d_llReaders.size(); ++i){
			FillDataFromBytes(&(buffer[2 * i]),d);
			d_llReaders[i]->UpdateFromData(d);
		}
	}


	void Device::UnlockRegister(Register::Address address){
		Register * reg = d_registers[address];
		if(! (reg->Flags() & (Register::WRITABLE | Register::PERSISTENT))){
			std::ostringstream os;
			os << "0x" << std::hex << reg->Flags() << " against " << "0x" << (int) (Register::WRITABLE | Register::PERSISTENT);
			throw std::logic_error("You can only unlock WRITABLE && PERSISTENT registers (" + os.str() + ")");
		}

		if(d_bus.CurrentWorkflow() != Bus::LAZY){
			throw std::runtime_error("Unlocking a register could only happen in Lazy workflow");
		}

		LazyWorkflow & w = d_bus.Lazy();

		Byte * out = w.Prepare(this->Klass(),this->ID(),UNLOCK_REGISTER,1);

		out[0] = address;

		const Byte * res(NULL);
		PayloadSize resSize;

		w.Submit(&res,resSize);

		w.Release();
		if( resSize != 0){
			std::ostringstream os;
			ADD_DEVICE_HEADER(this,os,address);
			os << " got " << resSize << " bytes but none is expected.";
			throw std::runtime_error(os.str());
		}
		return;

	}

	void Device::LowLatencyIOAccessor::ReadLowLatencyBuffer(Device & d,
	                                                        const Byte * buffer,
	                                                        PayloadSize size){
		d.ReadLowLatencyBuffer(buffer,size);
	}

	void Device::LowLatencyIOAccessor::WriteLowLatencyBuffer(Device & d,
	                                                         Byte * buffer,
	                                                         PayloadSize size){
		d.WriteLowLatencyBuffer(buffer,size);
	}


	void Device::EnablesLowLatencyPacket(const ListOfWriters & inputs,
	                                     const ListOfReaders & outputs){
		if(d_comVersion > SBCPVersion(0,1)){
			GetLowLatencyPacketDefinition();
			return;
		}
		d_llWriters = inputs;
		d_llReaders = outputs;
		d_maxllInputSize = inputs.size();
		d_maxllOutputSize = outputs.size();
	}

	void Device::GetLowLatencyPacketDefinition() {
		if (d_comVersion < SBCPVersion(0,2)) {
			//will not work
			return;
		}

		//we should be in LazyWorkflow;
		if(d_bus.CurrentWorkflow() != Bus::LAZY){
			throw std::logic_error("Getting low latency packet definition could only happen in Lazy workflow");
		}

		LazyWorkflow & w = d_bus.Lazy();

		w.Prepare(Klass(),ID(),Device::GET_LOW_LATENCY_PCK_DEF,0);

		const Byte * res(NULL);
		PayloadSize resSize;
		w.Submit(&res,resSize);


		if(resSize < 4) {
			w.Release();
			throw std::runtime_error("Unexpected size on Get Low Latency Packet Definition definition");
		}

		d_maxllInputSize   = res[0];
		d_maxllOutputSize  = res[1];
		const Byte inSize  = res[2];
		const Byte outSize = res[3];

		if(resSize < 4 + inSize + outSize) {
			w.Release();
			throw std::runtime_error("Unexpected size on Get Low Latency packet Definition definition");
		}

		d_llWriters.reserve(inSize);
		d_llReaders.reserve(outSize);

		//sets the inputs
		for(unsigned int i = 4; i < 4 + inSize; ++i){
			Register::Address adr = res[i];
			if(adr >= d_registers.size() || d_registers[adr] == NULL){
				std::ostringstream os;
				os << "Device low Latency Packet Definition specify register 0x"
				   << std::hex << (int) adr << " but it is unknown to our driver";
				throw std::runtime_error(os.str());
			}
			Register * reg = d_registers[adr];
			if( !(reg->Flags() & Register::WRITABLE) ) {
				std::ostringstream os;
				os << "Device low Latency packet definition specify input register 0x"
				   << std::hex << (int) adr << " but it is not writable";
				throw std::runtime_error(os.str());
			}
			//do not like dynamic cast, but this is needed here, and at configuration time
			d_llWriters.push_back(dynamic_cast<RegisterWriterBase *>(reg));
		}

		//sets the outputs
		for(unsigned int i = 4 + inSize; i < 4 + inSize + outSize; ++i){
			Register::Address adr = res[i];
			if(adr >= d_registers.size() || d_registers[adr] == NULL){
				std::ostringstream os;
				os << "Device low Latency packet definition specify register 0x"
				    << std::hex << (int) adr << " but it is unknown to our driver";
				throw std::runtime_error(os.str());
			}
			Register * reg = d_registers[adr];
			if( ~(reg->Flags()) & (Register::READABLE | Register::LL_READABLE) ) {
				std::ostringstream os;
				os << "Device low Latency packet definition specify output register 0x"
				   << std::hex << (int) adr << " but it is not readable and low"
				   << " latency readble";
				throw std::runtime_error(os.str());
			}

			//do not like dynamic cast, but this is needed here, and at configuration time
			d_llReaders.push_back(dynamic_cast<RegisterReaderBase *>(reg));
		}

		w.Release();

	}

	void Device::SetLowLatencyPacketDefinition(const ListOfWriters & inputs,
	                                           const ListOfReaders & outputs){
		if (d_comVersion < SBCPVersion(0,2)){
			std::ostringstream os;
			os << "Unavailable feature : The device communication version is "
			   << d_comVersion
			   << "Dynamic Low Latency Packet Definition have been introduce in version "
			   << SBCPVersion(0,2) << ".";
			throw std::runtime_error(os.str());
		}

		if(inputs.size() > d_maxllInputSize){
			std::ostringstream os;
			os << "Mismatched size for inputs, device accepts a maximum of "
			   << (int) d_maxllInputSize << " and user specified "
			   << inputs.size();
			throw std::overflow_error(os.str());
		}

		if(outputs.size() > d_maxllOutputSize){
			std::ostringstream os;
			os << "Mismatched size for outputs, device accepts a maximum of "
			   << (int) d_maxllOutputSize << " and user specified "
			   << outputs.size();
			throw std::overflow_error(os.str());
		}

		CheckWritersOwnership(inputs);
		CheckReaderOwnership(outputs);

		for(ListOfReaders::const_iterator r = outputs.begin();
		    r != outputs.end();
		    ++r){
			if( !( (*r)->Flags() & Register::LL_READABLE ) ) {
				std::ostringstream os;
				os << "Register 0x" << std::hex << (int) (*r)->GetAddress()
				   << " is not Register::LL_READABLE.";
				throw std::runtime_error(os.str());
			}
		}

		//we should be in LazyWorkflow;
		if(d_bus.CurrentWorkflow() != Bus::LAZY){
			throw std::logic_error("Setting low latency packet definition could only happen in Lazy workflow");
		}

		d_llReaders.clear();
		d_llWriters.clear();

		LazyWorkflow & w = d_bus.Lazy();

		Byte * out = w.Prepare(Klass(),
		                       ID(),
		                       SET_LOW_LATENCY_PCK_DEF,
		                       2 + inputs.size() + outputs.size());
		out[0] = inputs.size();
		out[1] = outputs.size();

		for(unsigned int i = 0; i < inputs.size(); ++i){
			out[2 + i] = inputs[i]->GetAddress();
		}

		for(unsigned int i = 0;
		    i < outputs.size();
		    ++i){
			out[2 + i + inputs.size()] = outputs[i]->GetAddress();
		}

		const Byte * res(NULL);
		PayloadSize size;
		w.Submit(&res,size);

		w.Release();
		d_llWriters = inputs;
		d_llReaders = outputs;
	}

	void Device::CheckReaderOwnership(const ListOfReaders & outputs) {
		for(ListOfReaders::const_iterator r = outputs.begin();
		    r != outputs.end();
		    ++r) {
			Register::Address adr = (*r)->GetAddress();
			if(adr >= d_registers.size() || d_registers[adr] == NULL){
				std::ostringstream os;
				os << "Device does not have a register 0x" << std::hex
				   << (int) adr << ".";
				throw std::logic_error(os.str());
			}

			Register * userReg = dynamic_cast<Register *>(*r);
			if(d_registers[adr] != userReg) {
				std::ostringstream os;
				os << "RegisterAccessor given by user ( " << userReg
				   << " ) do not match device intern object ( "
				   << d_registers[adr] << ")";
				throw std::logic_error(os.str());
			}
		}
	}

	void Device::CheckWritersOwnership(const ListOfWriters & inputs) {
		for(ListOfWriters::const_iterator w = inputs.begin();
		    w != inputs.end();
		    ++w) {
			Register::Address adr = (*w)->GetAddress();
			if(adr >= d_registers.size() || d_registers[adr] == NULL){
				std::ostringstream os;
				os << "Device does not have a register 0x" << std::hex
				   << (int) adr << ".";
				throw std::logic_error(os.str());
			}

			Register * userReg = dynamic_cast<Register *>(*w);
			if(d_registers[adr] != userReg) {
				std::ostringstream os;
				os << "RegisterAccessor given by user ( " << userReg
				   << " ) do not match device intern object ( "
				   << d_registers[adr] << ")";
				throw std::logic_error(os.str());
			}
		}
	}

	void Device::CheckSchedulable() throw (std::exception){
		if(!CanHaveALowLatencyPacket()) {
			throw std::runtime_error("Device is not schedulable for low latency as it does not support them.");
		}

		if(d_llReaders.size() == 0 && d_llWriters.size() == 0) {
			throw std::runtime_error("Device low latency packet is not configured, please use Device::SetLowLatencyPacket()");
		}
	}

	bool Device::IsSchedulable() {
		try{
			CheckSchedulable();
		} catch (const std::exception & ) {
			return false;
		}
		return true;
	}

	PayloadSize Device::LowLatencyOutSize() const {
		return d_llWriters.size() * 2;
	}

	PayloadSize Device::LowLatencyInSize() const {
		return d_llReaders.size() * 2;
	}

	unsigned int Device::MaxLowLatencyInSize() const{
		return d_maxllInputSize;
	}
	unsigned int Device::MaxLowLatencyOutSize() const{
		return d_maxllOutputSize;
	}

}/* namespace sbcp */
