/**
 * \file device/RegisterAccessor.h
 *
 *  \date Apr 30, 2012
 *  \author Alexandre Tuleu
 */

#ifndef LIBSBCP_DEVICE_REGISTERACCESSOR_H_
#define LIBSBCP_DEVICE_REGISTERACCESSOR_H_

#include "Device.h"

#include "RegisterAccessorBase.h"

/**
 *
 */
namespace sbcp{


/**
 * Base class for RegisterReader, RegisterWriter and RegisterAccessor.
 *
 * \tparam T : the type of the data for the user. The data of the register will
 *             automatically be converted to that type.
 * \tparam InpuType the input type of the data. Should be 16-bit bit size.
 *         Most ofen it is just a 16-bit int, signed or unsigned.

 */
template <typename T, typename InputType = sbcp::Data>
class RegisterAccessorBase {
public:
	RegisterAccessorBase(Device  & dev);
	virtual ~RegisterAccessorBase();

	/**
	 * @return last writen or read value.
	 */
	T LastValue() const;

	static void CheckFlags(unsigned int flg);

	const static Register::Flag AdditionnalFlagMask = (Register::Flag)( Register::PERSISTENT | Register::LL_READABLE);
protected :

	Device &                d_dev;
	T d_value;
};


/**
 * An object that could read a value of a Device register over the bus.
 *
 * They are usually created by Device class implementation constructor.
 * A end-user should not create them.
 *
 * They provide read access to the data through Get().
 *
 *
 */
template <typename T, typename InputType = sbcp::Data>
class RegisterReader : public virtual RegisterAccessorBase<T>,
                       public RegisterReaderBase {
public :
	typedef std::tr1::shared_ptr<RegisterReader<T,InputType> > Ptr;
	RegisterReader(Device & dev,
	               Register::Address address,
	               unsigned int additionalFlags = 0);
	virtual ~RegisterReader();

	/**
	 * @return A new Read of the value. Will perform an hardware tansfer if in
	 *  LazyWorkflow, so the thread may be blocked for a long time. In
	 *  ScheduledWorkflow returns the most updated value.
	 */
	T Get();


protected :
	virtual T FromInternal(InputType value);

private :
	typedef RegisterAccessorBase<T> Base;

	static Flag MergeFlags(unsigned int flg);

	void UpdateFromData(Data d);
};

/**
 * An object that could write to a Device Register.
 *
 * They are usually created by Device class implementation constructor.
 * A end-user should not create them.
 *
 * They provide write access to the data through Set().
 */
template <typename T, typename InputType = sbcp::Data>
class RegisterWriter : public virtual RegisterAccessorBase<T>,
                       public RegisterWriterBase {
public :
	typedef std::tr1::shared_ptr<RegisterWriter<T,InputType> > Ptr;
	RegisterWriter(Device & dev, Register::Address address,
	               unsigned int additionalFlags = 0);
	virtual ~RegisterWriter();

	/**
	 * Write the value to the Register. In LazyWorkflow, it will perform a
	 * hardware transfer, so the calling thread may be blocked. Otherwise the
	 * internal data is updated but no transfer performs.
	 * @param value The new value to write.
	 */

	void Set(T value);

	/**
	 * Unlocks the register for writing to a persistent register.
	 */
	void Unlock();

protected :
	virtual InputType FromExternal(T value);

private :
	typedef RegisterAccessorBase<T> Base;

	static Flag MergeFlags(unsigned int flg);

	InputType ToData();
};


/**
 * An object that could read and write to a Device register.
 *
 * They are usually created by Device class implementation constructor.
 * A end-user should not create them.
 *
 * I/O could be performed using Set() and Get().
 */
template <typename T, typename InputType = sbcp::Data>
class RegisterAccessor : public RegisterWriter<T>, public RegisterReader<T>{
public :
	typedef std::tr1::shared_ptr<RegisterAccessor<T,InputType> > Ptr;
	RegisterAccessor(Device & dev,
	                 Register::Address address,
	                 unsigned int additionalFlags = 0);
	virtual ~RegisterAccessor();

private :
	static Register::Flag MergeFlags(unsigned int flg);
};


} //namesapce sbcp

#endif // LIBSBCP_DEVICE_REGISTERACCESSOR_H_


#include "RegisterAccessor.impl.h"
