/*
 * \file RegisterAccessor.impl.h
 *
 *  \date Jun 3, 2013
 *  \author Alexandre Tuleu
 */

#ifndef LIBSBCP_DEVICE_REGISTERACCESSOR_IMPL_H_
#define LIBSBCP_DEVICE_REGISTERACCESSOR_IMPL_H_

#include "RegisterAccessor.h"
#include <libsbcp/bus/Bus.h>

#include <sstream>

namespace sbcp{

template <typename T, typename InpuType>
inline RegisterAccessorBase<T,InpuType>::RegisterAccessorBase(Device & dev)
	: d_dev(dev)
	, d_value(){
}



template <typename T, typename InpuType>
inline RegisterAccessorBase<T,InpuType>::~RegisterAccessorBase(){}



template <typename T, typename InpuType>
T RegisterAccessorBase<T,InpuType>::LastValue() const{
	return d_value;
}

template <typename T, typename InpuType>
void RegisterAccessorBase<T,InpuType>::CheckFlags(unsigned int flg){
	if(flg &(Register::READABLE | Register::WRITABLE)) {
		std::ostringstream os;
		os << "User defined flags ( 0x" << std::hex << flg
		   << " ) sets either Readable ( 0x" << Register::READABLE
		   << " ) or Writable ( 0x" << Register::WRITABLE
		   << " ) but they are manually managed.";
		throw std::logic_error(os.str());
	}
}

//RegisterReader


template <typename T, typename InputType>
inline RegisterReader<T,InputType>::RegisterReader(Device  & dev,
                                                   Register::Address address,
                                                   unsigned int additionalFlags)
	: Register(dev,
	           address,
	           MergeFlags(additionalFlags))
	, RegisterAccessorBase<T>(dev)
	, RegisterReaderBase(dev,address,MergeFlags(additionalFlags)){
}



template <typename T, typename InpuType>
inline RegisterReader<T,InpuType>::~RegisterReader(){}



template <typename T, typename InputType>
inline T RegisterReader<T,InputType>::Get(){
	if(Base::d_dev.GetBus().CurrentWorkflow() == Bus::SCHEDULED){
		return Base::d_value;
	}


	Base::d_value = FromInternal( Base::d_dev.GetRegister(GetAddress()));
	return Base::d_value;
}



template <typename T, typename InputType>
inline T RegisterReader<T,InputType>::FromInternal(InputType value){
	return (T) value;
}

template <typename T, typename InputType>
inline Register::Flag RegisterReader<T,InputType>::MergeFlags(unsigned int flg){
	Base::CheckFlags(flg);
	return (Register::Flag) (Register::READABLE | flg);
}

template <typename T, typename InputType>
inline void RegisterReader<T,InputType>::UpdateFromData(Data d){
	Base::d_value = FromInternal((InputType)d);
}


//RegisterWriter


template <typename T, typename InputType>
inline RegisterWriter<T,InputType>::RegisterWriter(Device  & dev,
                                                   Register::Address address,
                                                   unsigned int additionalFlags)
	: RegisterAccessorBase<T>(dev)
	, Register(dev,address,MergeFlags(additionalFlags))
	, RegisterWriterBase(dev,address,MergeFlags(additionalFlags)){
}



template <typename T, typename InputType>
inline RegisterWriter<T,InputType>::~RegisterWriter(){}



template <typename T, typename InputType>
inline void RegisterWriter<T,InputType>::Set(T value){
	Base::d_value = value;

	if(Base::d_dev.GetBus().CurrentWorkflow() == Bus::SCHEDULED ){
		return;
	}

	Base::d_dev.SetRegister(GetAddress(),FromExternal(Base::d_value));
}

template <typename T, typename InputType>
inline void RegisterWriter<T,InputType>::Unlock() {
	Base::d_dev.UnlockRegister(GetAddress());
}



template <typename T, typename InputType>
inline InputType RegisterWriter<T,InputType>::ToData(){
	return FromExternal(Base::d_value);
}

template <typename T, typename InputType>
inline Register::Flag RegisterWriter<T,InputType>::MergeFlags(unsigned int flg){
	Base::CheckFlags(flg);
	return (Register::Flag) (Register::WRITABLE | flg);
}

template <typename T, typename InputType>
inline InputType RegisterWriter<T,InputType>::FromExternal(T value){
	return (Data)value;
}


//RegisterAccessor


template <typename T, typename InputType>
inline RegisterAccessor<T,InputType>::RegisterAccessor(Device  & dev,
                                                       Register::Address address,
                                                       unsigned int additionalFlags)
	: Register(dev,
	           address,
	           MergeFlags(additionalFlags))
	, RegisterAccessorBase<T>(dev)
	, RegisterWriter<T>(dev,address,additionalFlags)
	, RegisterReader<T>(dev,address,additionalFlags)
{
}

template <typename T, typename InputType>
inline RegisterAccessor<T,InputType>::~RegisterAccessor(){}

template <typename T, typename InputType>
inline Register::Flag RegisterAccessor<T,InputType>::MergeFlags(unsigned int flg){
	RegisterAccessorBase<T,InputType>::CheckFlags(flg);
	return (Register::Flag) (Register::WRITABLE | Register::READABLE | flg);
}

} //namespace sbcp

#endif /* LIBSBCP_DEVICE_REGISTERACCESSOR_IMPL_H_ */
