/**
 * \file MotorDriver.h
 *
 *  \date Apr 10, 2012
 *  \author Alexandre Tuleu
 */

#ifndef LIBSBCP_MOTORDRIVER_H_
#define LIBSBCP_MOTORDRIVER_H_

#include <libsbcp/device/Device.h>
#include <libsbcp/device/RegisterAccessor.h>


namespace sbcp {
/**
 * namespace for all amarsi devices implementation.
 *
 * It holds currently :
 * MotorDriver : AMARSi MotorDriver board
 * Power : AMARSi power management board
 */
namespace amarsi {

	/**
	 * AMARSi motordriver board device driver.
	 *
	 * A motordriver is a SBCP device that drives :
	 * - two brushless Motor accessible with Motor1() and Motor2()
	 * - three MagneticEncoder accessible with Q1(), Q2(), Q3() and Q().
	 * - a three axis force sensor, accessible with Force()
	 *
	 * There is a close relation between Motor and MagneticEncoder. The later
	 * are used to runtime calibrate the motor. Please make sure before to do
	 * do anything with a Motor, to CalibrateMotors() if they are not
	 * Motor::RuntimeCalibrated(). Otherwise you may end up with a lot of
	 * TransferError because lot of changes of DeviceRegister like
	 * Motor::MotorControlMode() are forbidden if not
	 * Motor::RuntimeCalibrated(). Motor::CalibrateMotors() will return after
	 * the transfer, but you have to check manually that the Motor is
	 * Motor::RuntimeCalibrated() : Sometime it need to slowly move the motor,
	 * and it could take up to a few second to do so.
	 */
	class MotorDriver : public Device {
	public :
		/**
		 * helper typedef
		 */
		typedef std::tr1::shared_ptr<MotorDriver> Ptr;

		/**
		 * A class to drive a 1D brushless motor.
		 * Note that most of the data is 12-bit wide
		 * \todo : descriobe extensively the calibration procedure.
		 */
		class Motor {
		public :
			typedef std::tr1::shared_ptr<Motor> Ptr;
			/**
			 * Status for a motor.
			 */
			enum Status {
				IDLE               = 0,     //!< IDLE freshly started
				MOVING             = 1 << 0,//!< MOVING is moving
				RUNTIME_CALIBRATED = 1 << 1,//!< RUNTIME_CALIBRATED is runtime calibrated
				ERROR              = 1 << 2 //!< ERROR has an error
			};

			/**
			 * Available control mode for a motor.
			 */
			enum ControlMode {
				COAST               = 0,//!< COAST motor is free to move
				BRAKE               = 1,//!< BRAKE short cut of the winding.
				                        //!< Could break the electronic / motor
				                        //!< if there is a lot of force feedback.
				                        //!< Use with caution if not at all.
				POSITION            = 2,//!< POSITION
				VELOCITY            = 3,//!< VELOCITY
				SMOOTH_POSITION     = 4,//!< SMOOTH_POSITION
				TORQUE              = 5,//!< TORQUE
				RUNTIME_CALIBRATION = 6 //!< RUNTIME_CALIBRATION special mode,
				                        //!< that is setted only through
				                        //!< CalibrateMotors()
			};

			/**
			 * Available register for a Motor. Prefer to use the RgeisterAccessor()
			 */
			enum Register {
				STATUS_AND_CONTROL_MODE = 0,//!< STATUS_AND_CONTROL_MODE
				GOAL_POSITION,              //!< GOAL_POSITION
				TARGETED_POSITION,          //!< TARGETED_POSITION
				PRESENT_POSITION,           //!< PRESENT_POSITION

				GOAL_SPEED,                 //!< GOAL_SPEED
				TARGETED_SPEED,             //!< TARGETED_SPEED
				PRESENT_SPEED,              //!< PRESENT_SPEED

				TARGETED_TORQUE,            //!< TARGETED_TORQUE

				P_GAIN,                     //!< P_GAIN
				I_GAIN,                     //!< I_GAIN
				D_GAIN,                     //!< D_GAIN

				COMPLIANCE_PRELOAD,         //!< COMPLIANCE_PRELOAD
				COMPLIANCE_STIFFNESS,       //!< COMPLIANCE_STIFFNESS
				COMPLIANCE_DAMPING,         //!< COMPLIANCE_DAMPING

				MAX_ACCELERATION,           //!< MAX_ACCELERATION
				MAX_SPEED,                  //!< MAX_SPEED
				MAX_TORQUE,                 //!< MAX_TORQUE

				GEAR_MULT,                  //!< GEAR_MULT
				GEAR_DIV,                   //!< GEAR_DIV

				LIMIT_POS_CW,               //!< LIMIT_POS_CW
				CALIBRATION_REGION_MIN,     //!< CALIBRATION_REGION_MIN
				CALIBRATION_REGION_MAX,     //!< CALIBRATION_REGION_MAX
				CALIBRATION_OFFSET,         //!< CALIBRATION_OFFSET

				SMOOTH_POSITION_UPDATE,     //!< SMOOTH_POSITION_UPDATE

				MISSED_CONTROL_POINT,
				MAX_CONSECUTIVE_MISSED_CONTROL_POINT,
				OVERFILLED_BUFFER_CONTROL_STEP,
				UNUSED_MOTOR_REGISTER_START,

				REGISTER_SIZE = 32          //!< REGISTER_SIZE

			};

			Motor(Device & dev,
			      sbcp::Register::Address start);

			virtual  ~Motor();


			/**
			 * \name Position, Velocity and Torque Accessor
			 * List of RegisterAccessor / RegisterReader for goal/actual
			 * position,velocity and torque.
			 */
			///@{
			/**
			 * @return a RegisterAccessor for the Goal position.
			 */
			RegisterAccessor<double,int16_t>   & GoalPosition();
			/**
			 * @return a RegisterAccessor for the Goal velocity.
			 */
			RegisterAccessor<double,int16_t>   & GoalVelocity();

			/**
			 * @return a RegisterReader for the actual position of the motor.
			 * Current value measured from relative sensors.
			 */
			RegisterReader<double,int16_t> & PresentPosition();

			/**
			 * @return a RegisterReader for the actual velocity of the motor.
			 *  Current value measured from relative sensors
			 */
			RegisterReader<double,int16_t> & PresentVelocity();

			/**
			 * @return a RegisterReader for the current maximally allowed
			 * output torque of the motor. This is not the actual motorm output
			 * torque, only a maximal allowed value.
			 */
			RegisterReader<double> & TargetedTorque();
			///@}
			/**
			 * \name Control Accessor of the motor
			 *
			 * These are value that drive the coontrol algorithm of the motor:
			 * PID controller + virtual compliance
			 */
			///@{
			/**
			 * P gain of the PID.
			 * @return RegisterAccessor
			 */
			RegisterAccessor<Data> & PGain();
			/**
			 * I gain of the PID.
			 * @return RegisterAccessor
			 */
			RegisterAccessor<Data> & IGain();
			/**
			 * D gain of the PID.
			 * @return RegisterAccessor
			 */
			RegisterAccessor<Data> & DGain();

			/**
			 * Stiffness of the virtual compliance.
			 * @return RegisterAccessor
			 */
			RegisterAccessor<Data> & Stiffness();
			/**
			 * Damping of the virtual compliance.
			 * @return RegisterAccessor
			 */
			RegisterAccessor<Data> & Damping();
			/**
			 * Preload of the virtual compliance.
			 * @return RegisterAccessor
			 */
			RegisterAccessor<int16_t> & Preload();

			/**
			 * Maximal output torque.
			 * @return RegisterAccessor
			 */
			RegisterAccessor<Data> & MaxTorque();
			/**
			 * Maximal output speed.
			 * @return
			 */
			RegisterAccessor<Data> & MaxSpeed();
			/**
			 * maximal acceleration.
			 * @return
			 */
			RegisterAccessor<Data> & MaxAcceleration();
			///@}
			/**
			 *  \name Control accessor for a motor
			 *  RegisterAccessor to drive a motor
			 */
			///@{
			/**
			 * @return an RegisterAccessor for the ControlMode.
			 */
			RegisterAccessor<ControlMode> & MotorControlMode();
			/**
			 * Used only in SMOOTH_POSITION : refresh rate for the smooth
			 * position update.
			 * @return RegisterAccessor to the smooth refresh expected timestep in ms.
			 */
			RegisterAccessor<Data> & SmoothPositionUpdate();

			/**
			 * Multiplication of the Gear. Persistent
			 * @return RegisterAccessor
			 */
			RegisterAccessor<Data> & GearMult();

			/**
			 * Divisor of the Gear. Persistent
			 * @return RegisterAccessor
			 */
			RegisterAccessor<Data> & GearDiv();
			/**
			 * Maximal allowed position of the output shaft + static calibration status. Persistent.
			 *
			 * * 15 LSB : maximal position of the output shaft, should be in
			 *            ] 0 4096 [.
			 * * 1 MSB : 1 if statically calibrated (RegionMin() and RegionMax()
			 *           should be set carefully from the hardware),
			 * @return RegisterAccessor
			 */
			RegisterAccessor<Data> & PositionLimit();
			/**
			 * Minima of the calibration Region. Persistent.
			 * @return RegisterAccessor
			 */
			RegisterAccessor<Data> & RegionMin();
			/**
			 * Maxima of the calibration Region. Persistent.
			 * @return RegisterAccessor
			 */
			RegisterAccessor<Data> & RegionMax();
			/**
			 * Calibration offset.
			 * @return RegisterOffset.
			 */
			RegisterAccessor<int16_t> & CalibrationOffset();
			///@}
			/**
			 * Test if te motor is runtime calibrated.
			 * @return true if this is the case.
			 */

			RegisterReader<Data> & MissedControlPoint();
			RegisterReader<Data> & MaxConsecutiveMissedControlPoint();
			RegisterReader<Data> & OverfilledBufferControlStep();

			bool RuntimeCalibrated();

		private :


			RegisterAccessor<ControlMode>    d_mode;

			RegisterAccessor<double,int16_t> d_goalPosition;
			RegisterAccessor<double,int16_t> d_goalVelocity;

			RegisterReader<double,int16_t>  d_presentPosition;
			RegisterReader<double,int16_t>  d_presentVelocity;
			RegisterReader<double>          d_targetedTorque;

			RegisterAccessor<Data>  d_pgain;
			RegisterAccessor<Data>  d_igain;
			RegisterAccessor<Data>  d_dgain;


			RegisterAccessor<Data>  d_stiffness;
			RegisterAccessor<Data>  d_damping;
			RegisterAccessor<int16_t>  d_preload;

			RegisterAccessor<Data> d_MaxTorque;
			RegisterAccessor<Data> d_MaxAcceleration;
			RegisterAccessor<Data> d_MaxSpeed;

			RegisterAccessor<Data> d_gearMult,d_gearDiv,d_positionLimit,
			                            d_regionMin, d_regionMax;
			RegisterAccessor<int16_t> d_calibrationOffset;

			RegisterAccessor<Data> d_SmoothUpdate;

			RegisterReader<Data> d_missedControlPoint;
			RegisterReader<Data> d_maxMissedControlPoint;
			RegisterReader<Data> d_overfilledBufferStep;
		};
		/**
		 * Magnetic absolute Encoder driver
		 */
		class MagneticEncoder {
		public :
			typedef std::tr1::shared_ptr<MagneticEncoder> Ptr;
			enum Register {
				POSITION_AND_STATUS = 0,
				GEAR_MULT,
				GEAR_DIV,
				UNUSED_MAGNETIC_ENCODER,
				REGISTER_SIZE = 4
			};

			MagneticEncoder(Device & dev,
			                sbcp::Register::Address start);

			/**
			 * Position and status of the motor.
			 * - 15-14 MSB : status bits. if 0 evrything is fine, otherwise error.
			 * - 13-0 LSB : value of the encoder.
			 * @return RegisterReader
			 */
			RegisterReader<Data> & PositionAndStatus();
			/**
			 * Multiplier of the gear. Persistent
			 * @return RegisterAccessor,
			 */
			RegisterAccessor<int16_t> & GearMultiple();
			/**
			 * Dicisor of the gear. Persistent
			 * @return RegisterAccessor.
			 */
			RegisterAccessor<int16_t> & GearDivisor();

			uint16_t CurrentPosition();

			static bool IsGood(Data & positionAndStatus);

			static bool HasReadoutError(Data & positionAndStatus);

			static bool HasMagneticFieldError(Data & positionAndStatus);

		private :
			RegisterReader<Data> d_positionAndStatus;
			RegisterAccessor<int16_t> d_gearMult,d_gearDiv;
		};

		typedef RegisterReader<int16_t,int16_t> LoadCellReader;

		enum FORCE_AXIS_REGISTERS{
			FX_VALUE         = 0,
			FX_UNUSED_START     ,
			FX_REGISTER_SIZE = 4
		};
		enum Register {
			ME_Q1 = DEVICE_SPECIFIC_REGISTERS,                     //8  or 0x08
			ME_Q2 = ME_Q1 + MagneticEncoder::REGISTER_SIZE,        //12 or 0x0c
			ME_Q3 = ME_Q2 + MagneticEncoder::REGISTER_SIZE,        //16 or 0x10

			FORCE_AXIS_1 = ME_Q3 + MagneticEncoder::REGISTER_SIZE, //20 or 0x14
			FORCE_AXIS_2 = FORCE_AXIS_1 + FX_REGISTER_SIZE,        //24 or 0x18
			FORCE_AXIS_3 = FORCE_AXIS_2 + FX_REGISTER_SIZE,        //28 or 0x1c

			MOTOR_1 = FORCE_AXIS_3 + FX_REGISTER_SIZE,             //32 or 0x20
			MOTOR_2 = MOTOR_1 + Motor::REGISTER_SIZE               //64 or 0x40
		};

		enum Instruction {
			CALIBRATE        = DEVICE_SPECIFIC_INSTRUCTIONS,
			UNCALIBRATE      = DEVICE_SPECIFIC_INSTRUCTIONS + 1
		};
		const static sbcp::Class Klass = 0xea;
		MotorDriver(Bus & bus,
		            DevID id,
		            FirmwareVersion version,
		            SBCPVersion comVersion);
		virtual ~MotorDriver();

		/**
		 * @return a reference to the first Motor.
		 */
		Motor & Motor1();
		/**
		 * @return a reference to the second Motor.
		 */
		Motor & Motor2();

		/**
		 * @return a reference to the first absolute encoder.
		 */
		MagneticEncoder & Q1();
		/**
		 * @return a reference to the second absolute encoder.
		 */
		MagneticEncoder & Q2();
		/**
		 * @return a reference to the third absolute encoder.
		 */
		MagneticEncoder & Q3();
		/**
		 * @param i the index of the encoder to get in [0 2]
		 * @return reference to an absoilute encoder.
		 */
		MagneticEncoder & Q(unsigned int i);

		/**
		 * Runtime calibrates the motor. Switch to LazyWorkflow
		 */
		void CalibrateMotors();
		/**
		 * Runtime uncalibrate the motors.
		 */
		void UncalibrateMotors();

		LoadCellReader & Force(unsigned int dimension);


	private :
		Motor  d_m1, d_m2;
		MagneticEncoder::Ptr  d_q[3];
		LoadCellReader::Ptr  d_force[3];
	};

	inline MotorDriver::Motor & MotorDriver::Motor1(){
		return d_m1;
	}

	inline MotorDriver::Motor & MotorDriver::Motor2(){
		return d_m2;
	}


}
}

#endif // LIBSBCP_MOTORDRIVER_H_
