/**
 * \file TransferError.h
 *
 *  \date May 4, 2012
 *  \author Alexandre Tuleu
 */

#ifndef LIBSBCP_TRANSFERERROR_H_
#define LIBSBCP_TRANSFERERROR_H_

#include <stdexcept>

#include <vector>

#include <sbcpd/transfer.h>

#include <libsbcp/device/sbcp_common.h>


namespace sbcp{
/**
 * An error that happen duiring a physical tansfer. It could go from a non
 * responding device up to an application error on the device.
 *
 * They are fired when in LazyWorkflow() and performing a transfer, as when
 * RegisterAcessor::Set() or RegisterAccessor::Get()
 */
class TransferError : public std::runtime_error {
public:
	enum ErrorType {
		Application,
		Communication
	};
	TransferError(sbcpd_transfer * t);
	virtual ~TransferError() throw();
	virtual const char * what() const throw ();

	sbcp::Class DestinationClass() const;
	sbcp::DevID DestinationID() const;

	InstructionID Instruction() const;

	ErrorType  Type() const;

	sbcp::Class EmettorClass() const;
	sbcp::DevID EmettorID() const;

	Byte FeedbackByte() const;

	const std::vector<Byte> & PayloadData() const;


private :
	Class d_emettorClass, d_destClass;
	DevID d_emmettorID , d_destID;

	Byte        d_feedbackByteValue;
	sbcp::InstructionID d_instruction;
	ErrorType         d_type;
	std::vector<Byte> d_payloadData;

	std::string d_message;

};

/**
 * An error that could happend when performing several transfer at once
 * ( ScheduledWorkflow ). Mainly a holder for one or several TransferError.
 */

class MultipleTransferError : public std::runtime_error {
public :
	typedef std::vector<TransferError> ListOfError;
	MultipleTransferError();
	virtual ~MultipleTransferError() throw();

	virtual const char * what() const throw();

	const ListOfError & Errors() const;

	void AppendError(const TransferError & e);
private :
	ListOfError d_errors;
	std::string d_message;

};
}
#endif // LIBSBCP_TRANSFERERROR_H_
