/**
 * \file BusConfigSection.cpp
 *
 *  \date Dec 14, 2012
 *  \author Alexandre Tuleu
 */

#include "BusConfigSection.h"

#include <sbcpd/bus.h>
#include <sbcpd/si/ft2232h.h>
#include <sbcpd/si/ftdi.h>
#include <sbcpd/si/linux_serial.h>
namespace sbcp {

	BusConfigSection::SerialInterfaceConstructorByName BusConfigSection::s_siConstructors;

	BusConfigSection::BusConfigSection(options::Section::Ptr s)
		: options::ConfigSection(s)
		, d_hBaudrate(Section())
		, d_hSerialType(Section())
		, d_hFtdiInterface(Section())
		, d_hLatency(Section())
		, d_hDevName(Section())
		, d_hMasterID(Section())
		, d_hMinFrameSize(Section())
		, d_hMinPacketSize(Section()){
		InitSiConstructors();
	}

	BusConfigSection::~BusConfigSection(){
	}


	std::tr1::shared_ptr<Bus> BusConfigSection::OpenBus() const{
		return OpenBusWithFrame(MinFrameSize(),MinPacketSize());
	}

	std::tr1::shared_ptr<Bus> BusConfigSection::OpenBusWithFrame(unsigned int minFrameSize,
	                                                             unsigned int minPacketSize) const{
		typedef std::tr1::shared_ptr<Bus> BusPtr;

		sbcpd_bus_settings * s = sbcpd_bus_settings_create();

		sbcpd_bus_settings_set_required_frame_definition(s,
		                                                 minFrameSize,
		                                                 minPacketSize);

		SerialInterfaceConstructorByName::const_iterator
			fi = s_siConstructors.find(SerialType());
		if(fi == s_siConstructors.end()){
			throw std::runtime_error( "Could not find serial-type '"
			                         + SerialType() + "'");
		}

		sbcpd_serial_interface * si = fi->second->Create(*this);

		return BusPtr(new Bus(s,si));

	}

	SerialInterfaceConstructor::~SerialInterfaceConstructor(){}

	class FT2232HConstructor : public SerialInterfaceConstructor {
	public :
		virtual ~FT2232HConstructor(){}
		sbcpd_serial_interface * Create(const BusConfigSection & config){
			sbcpd_serial_interface * si = sbcpd_ft2232h_create_interface();
			if(sbcpd_ft2232h_open(si,
			                      config.Baudrate(),
			                      config.Latency(),
			                      config.FtdiInterface())){
				std::ostringstream os;
				os << "Error while opening ft2232h interface : "
				   << si->error.msg << ". Config was baudrate "
				   << config.Baudrate() << ", latency : "
				   << config.Latency() << " , interface : "
				   << config.FtdiInterface();
				sbcpd_ft2232h_close(si);
				throw std::runtime_error(os.str());
			}
			return si;
		}
	};

//	class FTDIConstructor : public SerialInterfaceConstructor {
//		virtual ~FT2232HConstructor(){}
//		sbcpd_serial_interface * Create(BusConfigSection & config){
//			sbcpd_serial_interface * si = sbcpd_ft2232h_create_interface();
//			if(sbcpd_ftdi_open(si,
//			                   config.Baudrate(),
//			                   config.Latency(),
//			                   config.FtdiInterface())){
//				std::ostringstream os;
//				os << "Error while opening ft2232h interface : "
//				   << si->error.msg << ". Config was baudrate "
//				   << config.Baudrate() << ", latency : "
//				   << config.Latency() << " , interface : "
//				   << config.FtdiInterface();
//				sbcpd_ft2232h_close(si);
//				throw std::runtime_error(os.str());
//			}
//			return si;
//		}
//	};

	class LinuxSerialConstructor : public SerialInterfaceConstructor {
	public :
		virtual ~LinuxSerialConstructor(){}
		sbcpd_serial_interface * Create(const BusConfigSection & config){
			sbcpd_serial_interface * si = sbcpd_linux_serial_create_interface();
			if(sbcpd_linux_serial_open(si,
			                           config.DevName().c_str(),
			                           config.Baudrate())){
				std::ostringstream os;
				os << "Error while opening linux serial interface : "
				   << si->error.msg << ". Config was baudrate "
				   << config.Baudrate() << ", device : '"
				   << config.DevName() << "'";
				sbcpd_linux_serial_close(si);
				throw std::runtime_error(os.str());
			}
			return si;
		}
	};

	void BusConfigSection::InitSiConstructors(){
		if(!s_siConstructors.empty()){
			return;
		}
		s_siConstructors["linux"] = SerialInterfaceConstructor::Ptr(new LinuxSerialConstructor());
		s_siConstructors["ft2232h"] = SerialInterfaceConstructor::Ptr(new FT2232HConstructor());
	}


} /* namespace sbcp */
